import os
from pathlib import Path, PurePath
from typing import Optional
from urllib.parse import quote as urlquote

from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import get_plugin_logger
from mkdocs.structure.files import File, Files

from mkdocs_static_i18n.config import RE_LOCALE

log = get_plugin_logger(__name__)


def create_i18n_file(
    file: File,
    current_language: str,
    default_language: str,
    all_languages: list,
    config: MkDocsConfig,
) -> File:
    log.debug(f"reconfigure {file}")

    # create a new File instance that we can turn into an i18n file
    file = File(
        file.src_path,
        config.docs_dir,
        file.dest_path,
        config.use_directory_urls,
        dest_uri=file.dest_uri,
        inclusion=file.inclusion,
    )

    file_dest_path = Path(file.dest_path)
    file_locale = default_language
    file_localization = None
    locale_site_dir = current_language if current_language != default_language else ""

    try:
        file_locale = Path(file.name).suffixes.pop(-1).replace(".", "")
        # the file_locale must be a valid language locale code that we check on the
        # configured languages (validated by config) or using the locale regex in case
        # users have localized files but not configured them on the plugin.languages yet
        if file_locale in all_languages or RE_LOCALE.match(file_locale):
            file_dest_path_no_suffix = Path(file_dest_path.parent) / Path(
                file_dest_path.stem
            ).with_suffix("")
            file_dest_path = Path(f"{file_dest_path_no_suffix}{file_dest_path.suffix}")
            # file name suffixed by locale, remove it
            file.name = Path(file.name).stem
            # store file localization
            file_localization = file_locale
        else:
            file_locale = default_language
    except IndexError:
        pass

    if config.use_directory_urls and file_localization:
        if file_dest_path.parent != Path("."):
            parent_dest_path_no_suffix = file_dest_path.parent.with_suffix("")
            file_dest_path = parent_dest_path_no_suffix / file_dest_path.name
            # named index.lang.md get transformed to index.lang.md/index.md
            # so we need to remove that named folder
            if file_dest_path.parent.name in ["index", "README"]:
                file_dest_path = file_dest_path.parent.parent / file_dest_path.name

    # README.html should be renamed to index.html
    if file_dest_path.name.endswith("README.html"):
        file_dest_path = file_dest_path.with_name("index.html")

    if locale_site_dir:
        file.dest_path = Path(locale_site_dir) / file_dest_path
    else:
        file.dest_path = file_dest_path
    file.abs_dest_path = os.path.normpath(os.path.join(config.site_dir, file.dest_path))

    # assure a valid name for the Page.is_index check
    if file.name == "README":
        file.name = "index"

    # update the url in case we played with the folder structure
    file.url = file._get_url(config.use_directory_urls)

    # fix the url
    if locale_site_dir:
        if PurePath(file.url).as_posix() == ".":
            file.url = urlquote(locale_site_dir + "/")

    # save some i18n metadata
    # alternates should list themselves
    file.alternates = {current_language: file}
    file.locale = file_locale
    file.locale_alternate_of = current_language
    file.localization = file_localization

    log.debug(f"reconfigure {file} from locale {file_locale}")

    return file


class I18nFiles(Files):
    """
    We need to override Files to handle the automagic discovery of resources allowing
    users to not specify their suffixed version of the resource (think image).

    Example: french page displaying image.fr.png can use [my french img](image.png)

    This is made possible by overriding 'get_file_from_path' below.
    """

    def __init__(self, plugin, files: Files) -> None:
        super().__init__(files)
        self.plugin = plugin

    def get_file_from_path(self, path: str) -> Optional[File]:
        """
        Used by mkdocs.structure.nav.get_navigation to find resources linked in markdown.
        """
        expected_src_uri = PurePath(path)
        if expected_src_uri == PurePath("."):
            expected_src_uri = PurePath("index.md")
        expected_src_uris = [
            expected_src_uri.with_suffix(
                f".{self.plugin.current_language}{expected_src_uri.suffix}"
            ),
            expected_src_uri.with_suffix(
                f".{self.plugin.default_language}{expected_src_uri.suffix}"
            ),
            expected_src_uri,
        ]
        for src_uri in expected_src_uris:
            file = self.src_uris.get(src_uri.as_posix())
            if file:
                return file
        else:
            return None



from mkdocs.commands.build import build
from mkdocs.config.base import load_config


def test_hooks_working():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "i18n": {
                "languages": [
                    {"locale": "en", "name": "english", "default": True},
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
        hooks=["hooks.py"],
    )
    build(mkdocs_config)


def test_hooks_env_modified():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs", "custom_dir": "theme_overrides"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "i18n": {
                "languages": [
                    {"locale": "en", "name": "english", "default": True},
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
        hooks=["hooks_jinja_on_env.py"],
        site_dir="/tmp/xoxo/",
    )
    build(mkdocs_config)



import os
from pathlib import Path, PurePath
from typing import Optional

from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import get_plugin_logger
from mkdocs.structure.files import File, Files

from mkdocs_static_i18n import is_relative_to
from mkdocs_static_i18n.config import RE_LOCALE

log = get_plugin_logger(__name__)


def create_i18n_file(
    file: File,
    current_language: str,
    default_language: str,
    all_languages: list,
    config: MkDocsConfig,
) -> File:
    log.debug(f"reconfigure {file}")

    # create a new File instance that we can turn into an i18n file
    file = File(
        file.src_path,
        config.docs_dir,
        file.dest_path,
        config.use_directory_urls,
        dest_uri=file.dest_uri,
        inclusion=file.inclusion,
    )

    file_dest_path = Path(file.dest_path)
    file_locale = default_language
    file_localization = None

    for language in all_languages:
        if is_relative_to(file.src_path, language):
            file_locale = language
            file_localization = language
            break
        else:
            # maybe the language folder is present BUT not configured in the plugin.languages yet
            # so we use the locale regex to guess that it's a localization folder
            maybe_file_locale = PurePath(file.src_path).parts[0]
            if RE_LOCALE.match(maybe_file_locale):
                file_locale = maybe_file_locale
                file_localization = maybe_file_locale
                break

    # README.html should be renamed to index.html
    if file_dest_path.name.endswith("README.html"):
        file_dest_path = file_dest_path.with_name("index.html")

    if file_locale != current_language:
        if is_relative_to(file_dest_path, file_locale):
            # we have to change the output folder
            file.dest_path = PurePath(
                PurePath(current_language) / PurePath(file_dest_path).relative_to(file_locale)
            ).as_posix()
        elif current_language != default_language:
            file.dest_path = PurePath(
                PurePath(current_language) / PurePath(file_dest_path)
            ).as_posix()
    elif file_locale == default_language:
        # we check that we are relative since we also accept other non localized files (assets)
        if is_relative_to(file_dest_path, file_locale):
            # we have to change the output folder to nothing
            file.dest_path = PurePath(file_dest_path).relative_to(file_locale).as_posix()
        elif current_language != default_language:
            file.dest_path = PurePath(
                PurePath(current_language) / PurePath(file_dest_path)
            ).as_posix()
    file.abs_dest_path = os.path.normpath(os.path.join(config.site_dir, file.dest_path))

    # assure a valid name for the Page.is_index check
    if file.name == "README":
        file.name = "index"

    # update the url in case we played with the folder structure
    file.url = file._get_url(config.use_directory_urls)

    # save some i18n metadata
    # alternates should list themselves
    file.alternates = {current_language: file}
    file.locale = file_locale
    file.locale_alternate_of = current_language
    file.localization = file_localization

    log.debug(f"reconfigure {file} from locale {file_locale}")

    return file


class I18nFiles(Files):
    """
    We need to override Files to handle the automagic discovery of resources allowing
    users to not specify their suffixed version of the resource (think image).

    Example: french page displaying image.fr.png can use [my french img](image.png)

    This is made possible by overriding 'get_file_from_path' below.
    """

    def __init__(self, plugin, files: Files) -> None:
        super().__init__(files)
        self.plugin = plugin

    def get_file_from_path(self, path: str) -> Optional[File]:
        """
        Used by mkdocs.structure.nav.get_navigation to find resources linked in markdown.
        """
        expected_src_uri = PurePath(path)
        if expected_src_uri == PurePath("."):
            expected_src_uri = PurePath("index.md")
        expected_src_uris = []
        # non localized paths detection (root)
        if is_relative_to(expected_src_uri, self.plugin.current_language):
            expected_src_uris.append(expected_src_uri.relative_to(self.plugin.current_language))
        elif is_relative_to(expected_src_uri, self.plugin.default_language):
            expected_src_uris.append(expected_src_uri.relative_to(self.plugin.default_language))
            if self.plugin.config.fallback_to_default is True:
                expected_src_uris.append(
                    PurePath(self.plugin.current_language)
                    / expected_src_uri.relative_to(self.plugin.default_language)
                )
        # localized paths detection
        else:
            expected_src_uris.append(PurePath(self.plugin.current_language) / expected_src_uri)
        # default language path detection
        if self.plugin.config.fallback_to_default is True:
            expected_src_uris.append(PurePath(self.plugin.default_language) / expected_src_uri)
        expected_src_uris.append(expected_src_uri)

        for src_uri in expected_src_uris:
            file = self.src_uris.get(src_uri.as_posix())
            if file:
                return file
        else:
            return None


def reconfigure_navigation(i18n_plugin, nav):
    """
    Remove the Section(title='En') language sections.
    """
    items = []
    for section in filter(lambda i: i.is_top_level and i.is_section, nav.items):
        if (
            section.title == i18n_plugin.current_language.capitalize()
            or section.title == i18n_plugin.default_language.capitalize()
        ):
            items.extend(section.children)
    if items:
        nav.items = items
        # [Page(title=[blank], url='/mkdocs-static-i18n/fr/'), Section(title='Topic1'), Section(title='Topic2'), Section(title='French only'), Section(title='English default')]
        # =>
        # [Page(title=[blank], url='/mkdocs-static-i18n/fr/'), Section(title='English default'), Section(title='French only'), Section(title='Topic1'), Section(title='Topic2')]
        nav.items.sort(key=lambda x: x.title if x.title else "")
    return nav



from re import compile

from mkdocs.config import config_options
from mkdocs.config.base import Config, ValidationError

RE_LOCALE = compile(r"(^[a-z]{2}(-[A-Za-z]{4})?(-[A-Z]{2})?$)|(^[a-z]{2}_[A-Z]{2}$)")


class Locale(config_options.Type):
    """
    Locale Config Option

    Validate the locale config option against a given Python type.
    """

    def run_validation(self, value):
        value = super().run_validation(value)
        if not RE_LOCALE.match(value) and value.lower() != "null":
            raise ValidationError(
                "Language code values must be either ISO-639-1 lower case "
                "or represented with their territory/region/country codes, "
                f"received '{value}' expected forms examples: 'en' or 'en-US' or 'en_US'."
            )
        return value


class I18nPluginLanguage(Config):
    """ """

    build = config_options.Type(bool, default=True)
    copyright = config_options.Optional(config_options.Type(str))
    default = config_options.Type(bool, default=False)
    extra = config_options.Optional(config_options.Type(dict))
    fixed_link = config_options.Optional(config_options.Type(str))
    link = config_options.Optional(config_options.Type(str))
    locale = Locale(str)
    name = config_options.Type(str)
    nav = config_options.Optional(config_options.Nav())
    nav_translations = config_options.Optional(config_options.Type(dict))
    site_author = config_options.Optional(config_options.Type(str))
    site_description = config_options.Optional(config_options.Type(str))
    site_name = config_options.Optional(config_options.Type(str))
    site_url = config_options.Optional(config_options.Type(str))
    theme = config_options.Optional(config_options.Type(dict))

    def validate(self):
        failed, warnings = super().validate()
        # set defaults
        if self.link is None:
            if self.default:
                self.link = "/"
            else:
                self.link = f"/{self.locale}/"
        # link should be absolute and end with a trailing /
        else:
            if not self.link.startswith("/") or not self.link.endswith("/"):
                failed.append(
                    (
                        "link",
                        ValidationError(
                            f"languages[{self.locale}].link should be an absolute link starting "
                            f"with a leading / and ending with a / (like /{self.locale}/)."
                        ),
                    )
                )
        # special "null" locale can be used to insert an item in the language switcher
        # so it must never be built
        if self.locale == "null":
            self.build = False
            if self.fixed_link is None:
                failed.append(
                    (
                        "fixed_link",
                        ValidationError(
                            f"languages[{self.locale}].fixed_link should be set to a valid URL."
                        ),
                    )
                )
        return failed, warnings


class I18nPluginConfig(Config):
    """ """

    docs_structure = config_options.Choice(["folder", "suffix"], default="suffix")
    fallback_to_default = config_options.Type(bool, default=True)
    reconfigure_material = config_options.Type(bool, default=True)
    reconfigure_search = config_options.Type(bool, default=True)
    languages = config_options.ListOfItems(
        config_options.SubConfig(I18nPluginLanguage, validate=True)
    )

    def validate(self):
        failed, warnings = super().validate()
        if not failed:
            # check that we have at least a default language to build
            if not any(filter(lambda c: c.default and c.build, self.languages)):
                failed.append(
                    (
                        "languages",
                        ValidationError(
                            "Could not find a default language to build "
                            "(expecting 'default: true' AND 'build: true')."
                        ),
                    )
                )
        return failed, warnings



import pytest
from mkdocs import utils
from mkdocs.config.base import load_config
from mkdocs.structure.files import get_files
from mkdocs.structure.nav import get_navigation


@pytest.fixture
def make_config():
    created_configs = []

    def _make_config(mkdocs_fp=None, docs_dir=None, use_directory_urls=True, plugins={}):
        config = load_config(
            mkdocs_fp,
            docs_dir=docs_dir,
            theme={"name": "mkdocs"},
            use_directory_urls=use_directory_urls,
            plugins=plugins,
        )
        config = config.plugins.run_event("config", config)
        files = get_files(config)
        files = config.plugins.run_event("files", files, config=config)
        env = config.theme.get_env()
        nav = get_navigation(files, config)
        nav = config.plugins.run_event("nav", nav, files=files, config=config)
        #
        created_configs.append(config)
        #
        return config, files, env, nav

    yield _make_config

    for config in created_configs:
        utils.clean_directory(config.site_dir)


@pytest.fixture
def make_localized_config(make_config):
    created_configs = []

    def _make_localized_config(config, locale):
        config.plugins["i18n"].current_language = locale
        #
        config = config.plugins.run_event("config", config)
        files = get_files(config)
        files = config.plugins.run_event("files", files, config=config)
        env = config.theme.get_env()
        nav = get_navigation(files, config)
        nav = config.plugins.run_event("nav", nav, files=files, config=config)
        #
        created_configs.append(config)
        #
        return config, files, env, nav

    yield _make_localized_config

    for config in created_configs:
        utils.clean_directory(config.site_dir)



import pytest
from mkdocs.commands.build import build


def navigate(nav):
    for item in nav:
        if item.children:
            for child in navigate(item.children):
                yield child
        yield item


def test_control_build(make_config):
    control_config, control_files, control_env, control_nav = make_config(
        "tests/structures/control/en_only/mkdocs.yml",
    )
    build(control_config)


@pytest.mark.parametrize(
    "control_data,test_data",
    [
        (
            {"mkdocs_fp": "tests/structures/control/en_only/mkdocs.yml"},
            {
                "mkdocs_fp": "tests/mkdocs.yml",
                "docs_dir": "docs_suffix_structure_one_language/",
                "plugins": {
                    "i18n": {
                        "languages": [
                            {"locale": "en", "name": "test", "default": True},
                        ],
                    }
                },
            },
        ),
        (
            {"mkdocs_fp": "tests/structures/control/fr_only/mkdocs.yml"},
            {
                "mkdocs_fp": "tests/mkdocs.yml",
                "docs_dir": "docs_suffix_structure_one_language/",
                "plugins": {
                    "i18n": {
                        "languages": [
                            {"locale": "fr", "name": "test", "default": True},
                        ],
                    }
                },
            },
        ),
        (
            {"mkdocs_fp": "tests/structures/control/en_only/mkdocs.yml"},
            {
                "mkdocs_fp": "tests/mkdocs.yml",
                "docs_dir": "docs_folder_structure_one_language/",
                "plugins": {
                    "i18n": {
                        "docs_structure": "folder",
                        "languages": [
                            {"locale": "en", "name": "test", "default": True},
                        ],
                    }
                },
            },
        ),
        (
            {"mkdocs_fp": "tests/structures/control/fr_only/mkdocs.yml"},
            {
                "mkdocs_fp": "tests/mkdocs.yml",
                "docs_dir": "docs_folder_structure_one_language/",
                "plugins": {
                    "i18n": {
                        "docs_structure": "folder",
                        "languages": [
                            {"locale": "fr", "name": "test", "default": True},
                        ],
                    }
                },
            },
        ),
    ],
)
def test_control_single(make_config, control_data, test_data):
    control_config, control_files, control_env, control_nav = make_config(**control_data)
    test_config, test_files, test_env, test_nav = make_config(**test_data)
    #
    assert control_nav.__str__() == test_nav.__str__()
    assert control_env.filters.keys() == test_env.filters.keys()
    assert {file.dest_uri for file in control_files} == {file.dest_uri for file in test_files}

    for control_page, test_page in zip(
        control_files.documentation_pages(), test_files.documentation_pages()
    ):
        assert control_page.dest_uri == test_page.dest_uri
        assert control_page.name == test_page.name
        assert control_page.url == test_page.url


@pytest.mark.parametrize(
    "control_data_en,control_data_fr,test_data",
    [
        (
            {"mkdocs_fp": "tests/structures/control/en_only/mkdocs.yml"},
            {"mkdocs_fp": "tests/structures/control/fr_with_default/mkdocs.yml"},
            {
                "mkdocs_fp": "tests/mkdocs.yml",
                "docs_dir": "docs_suffix_structure_two_languages/",
                "plugins": {
                    "i18n": {
                        "languages": [
                            {"locale": "en", "name": "test", "default": True},
                            {"locale": "fr", "name": "test"},
                        ],
                    }
                },
            },
        ),
        (
            {"mkdocs_fp": "tests/structures/control/en_only/mkdocs.yml"},
            {"mkdocs_fp": "tests/structures/control/fr_without_default/mkdocs.yml"},
            {
                "mkdocs_fp": "tests/mkdocs.yml",
                "docs_dir": "docs_suffix_structure_two_languages/",
                "plugins": {
                    "i18n": {
                        "fallback_to_default": False,
                        "languages": [
                            {"locale": "en", "name": "test", "default": True},
                            {"locale": "fr", "name": "test"},
                        ],
                    }
                },
            },
        ),
        (
            {"mkdocs_fp": "tests/structures/control/en_only/mkdocs.yml"},
            {"mkdocs_fp": "tests/structures/control/fr_with_default/mkdocs.yml"},
            {
                "mkdocs_fp": "tests/mkdocs.yml",
                "docs_dir": "docs_suffix_structure_two_languages/",
                "plugins": {
                    "i18n": {
                        "languages": [
                            {"locale": "en", "name": "test", "default": True},
                            {"locale": "fr", "name": "test"},
                            {"locale": "null", "name": "help", "build": True, "fixed_link": "https://ultrabug.fr"},
                        ],
                    }
                },
            },
        ),
        (
            {"mkdocs_fp": "tests/structures/control/en_only/mkdocs.yml"},
            {"mkdocs_fp": "tests/structures/control/fr_with_default/mkdocs.yml"},
            {
                "mkdocs_fp": "tests/mkdocs.yml",
                "docs_dir": "docs_folder_structure_two_languages_with_default/",
                "plugins": {
                    "i18n": {
                        "docs_structure": "folder",
                        "languages": [
                            {"locale": "en", "name": "test", "default": True},
                            {"locale": "fr", "name": "test"},
                        ],
                    }
                },
            },
        ),
        (
            {"mkdocs_fp": "tests/structures/control/en_only/mkdocs.yml"},
            {"mkdocs_fp": "tests/structures/control/fr_without_default/mkdocs.yml"},
            {
                "mkdocs_fp": "tests/mkdocs.yml",
                "docs_dir": "docs_folder_structure_two_languages/",
                "plugins": {
                    "i18n": {
                        "fallback_to_default": False,
                        "docs_structure": "folder",
                        "languages": [
                            {"locale": "en", "name": "test", "default": True},
                            {"locale": "fr", "name": "test"},
                        ],
                    }
                },
            },
        ),
        (
            {"mkdocs_fp": "tests/structures/control/en_only/mkdocs.yml"},
            {"mkdocs_fp": "tests/structures/control/fr_with_default/mkdocs.yml"},
            {
                "mkdocs_fp": "tests/mkdocs.yml",
                "docs_dir": "docs_folder_structure_two_languages_with_default/",
                "plugins": {
                    "i18n": {
                        "docs_structure": "folder",
                        "languages": [
                            {"locale": "en", "name": "test", "default": True},
                            {"locale": "fr", "name": "test"},
                            {"locale": "null", "name": "help", "build": True, "fixed_link": "https://ultrabug.fr"},
                        ],
                    }
                },
            },
        ),
    ],
)
def test_control_en_fr(
    make_config, make_localized_config, control_data_en, control_data_fr, test_data
):
    # default

    (
        control_config_en,
        control_files_en,
        control_env_en,
        control_nav_en,
    ) = make_config(**control_data_en)
    test_config, test_files, test_env, test_nav = make_config(**test_data)

    assert control_nav_en.__str__() == test_nav.__str__()
    assert control_env_en.filters.keys() == test_env.filters.keys()
    assert {file.dest_uri for file in control_files_en} == {file.dest_uri for file in test_files}

    for control_page, test_page in zip(
        control_files_en.documentation_pages(), test_files.documentation_pages()
    ):
        assert control_page.dest_uri == test_page.dest_uri
        assert control_page.name == test_page.name
        assert control_page.url == test_page.url

    # localized

    (
        control_config_fr,
        control_files_fr,
        control_env_fr,
        control_nav_fr,
    ) = make_config(**control_data_fr)
    (
        test_config_fr,
        test_files_fr,
        test_env_fr,
        test_nav_fr,
    ) = make_localized_config(test_config, "fr")
    #
    assert control_nav_fr.__str__() == test_nav_fr.__str__()
    assert control_env_fr.filters.keys() == test_env_fr.filters.keys()

    control_dest_uris = set()
    for file in control_files_fr:
        if not file.dest_uri.startswith("assets/"):
            control_dest_uris.add(f"fr/{file.dest_uri}")
        else:
            control_dest_uris.add(file.dest_uri)
    assert control_dest_uris == {file.dest_uri for file in test_files_fr}

    control_pages = control_files_fr.documentation_pages()
    control_pages.sort(key=lambda p: p.src_uri)
    test_pages = test_files_fr.documentation_pages()
    test_pages.sort(key=lambda p: p.src_uri)

    for control_page, test_page in zip(control_pages, test_pages):
        assert f"fr/{control_page.dest_uri}" == test_page.dest_uri
        assert control_page.name == test_page.name
        if control_page.url == "./":
            assert "fr/" == test_page.url
        else:
            assert f"fr/{control_page.url}" == test_page.url



from mkdocs.config.base import load_config

from mkdocs_static_i18n.plugin import I18n


def test_plugin_single_language_en():
    plugin = I18n()
    plugin.load_config({"languages": [{"locale": "en", "name": "english", "default": True}]})
    config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_one_language/",
    )
    result = plugin.on_config(config)
    assert str(result["theme"]["locale"]) == "en"


def test_plugin_single_language_fr():
    plugin = I18n()
    plugin.load_config({"languages": [{"locale": "fr", "name": "français", "default": True}]})
    config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_one_language/",
    )
    result = plugin.on_config(config)
    assert str(result["theme"]["locale"]) == "fr"


def test_plugin_theme_sitemap():
    plugin = I18n()
    plugin.load_config({"languages": [{"locale": "fr", "name": "français", "default": True}]})
    config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_one_language/",
    )
    result = plugin.on_config(config)
    assert result["theme"].dirs[0].endswith("custom_i18n_sitemap")



from pathlib import Path

from mkdocs.commands.build import build
from mkdocs.config.base import load_config

USE_DIRECTORY_URLS = [
    Path("english_default/index.en/index.html"),
    Path("404.html"),
    Path("assets/image_non_localized.png"),
    Path("french_only/index.fr/index.html"),
    Path("image.en.fake"),
    Path("image.en.png"),
    Path("image.fr.fake"),
    Path("image.fr.png"),
    Path("index.fr/index.html"),
    Path("index.html"),
    Path("topic1/named_file.en/index.html"),
    Path("topic1/named_file.fr/index.html"),
    Path("topic2/1.1.filename.fr.html"),
    Path("topic2/1.1.filename.html"),
    Path("topic2/index.html"),
    Path("topic2/README.fr/index.html"),
    Path("topic2/release_notes_17.1/index.html"),
    Path("topic2/release_notes_17.2/index.html"),
]
NO_USE_DIRECTORY_URLS = [
    Path("english_default/index.en.html"),
    Path("404.html"),
    Path("assets/image_non_localized.png"),
    Path("french_only/index.fr.html"),
    Path("image.en.fake"),
    Path("image.en.png"),
    Path("image.fr.fake"),
    Path("image.fr.png"),
    Path("index.fr.html"),
    Path("index.html"),
    Path("topic1/named_file.en.html"),
    Path("topic1/named_file.fr.html"),
    Path("topic2/1.1.filename.fr.html"),
    Path("topic2/1.1.filename.html"),
    Path("topic2/index.html"),
    Path("topic2/README.fr.html"),
    Path("topic2/release_notes_17.1.html"),
    Path("topic2/release_notes_17.2.html"),
]


def test_build_use_directory_urls():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
    )
    build(mkdocs_config)
    site_dir = mkdocs_config["site_dir"]
    generate_site = [f.relative_to(site_dir) for f in Path(site_dir).glob("**/*.html")]
    generate_site.extend([f.relative_to(site_dir) for f in Path(site_dir).glob("**/image*.*")])
    assert sorted(generate_site) == sorted(USE_DIRECTORY_URLS)


def test_build_no_use_directory_urls():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=False,
        docs_dir="docs_suffix_structure_two_languages/",
    )
    build(mkdocs_config)
    site_dir = mkdocs_config["site_dir"]
    generate_site = [f.relative_to(site_dir) for f in Path(site_dir).glob("**/*.html")]
    generate_site.extend([f.relative_to(site_dir) for f in Path(site_dir).glob("**/image*.*")])
    assert sorted(generate_site) == sorted(NO_USE_DIRECTORY_URLS)


PLUGIN_USE_DIRECTORY_URLS = [
    Path("english_default/index.html"),
    Path("404.html"),
    Path("image.png"),
    Path("image.fake"),
    Path("index.html"),
    Path("assets/image_non_localized.png"),
    Path("topic1/named_file/index.html"),
    Path("topic2/index.html"),
    Path("topic2/1.1.filename.html"),
    Path("topic2/release_notes_17.1/index.html"),
    Path("topic2/release_notes_17.2/index.html"),
    Path("fr/index.html"),
    Path("fr/image.png"),
    Path("fr/image.fake"),
    Path("fr/english_default/index.html"),
    Path("fr/topic1/named_file/index.html"),
    Path("fr/topic2/index.html"),
    Path("fr/topic2/1.1.filename.html"),
    Path("fr/french_only/index.html"),
    Path("fr/topic2/release_notes_17.1/index.html"),
    Path("fr/topic2/release_notes_17.2/index.html"),
]
PLUGIN_NO_USE_DIRECTORY_URLS = [
    Path("english_default/index.html"),
    Path("404.html"),
    Path("image.png"),
    Path("image.fake"),
    Path("index.html"),
    Path("assets/image_non_localized.png"),
    Path("topic1/named_file.html"),
    Path("topic2/index.html"),
    Path("topic2/1.1.filename.html"),
    Path("topic2/release_notes_17.1.html"),
    Path("topic2/release_notes_17.2.html"),
    Path("fr/index.html"),
    Path("fr/image.png"),
    Path("fr/image.fake"),
    Path("fr/english_default/index.html"),
    Path("fr/topic1/named_file.html"),
    Path("fr/topic2/index.html"),
    Path("fr/topic2/1.1.filename.html"),
    Path("fr/french_only/index.html"),
    Path("fr/topic2/release_notes_17.1.html"),
    Path("fr/topic2/release_notes_17.2.html"),
]


def test_plugin_use_directory_urls():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "search": {},
            "i18n": {
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
    )
    build(mkdocs_config)
    site_dir = mkdocs_config["site_dir"]
    generate_site = [f.relative_to(site_dir) for f in Path(site_dir).glob("**/*.html")]
    generate_site.extend([f.relative_to(site_dir) for f in Path(site_dir).glob("**/image*.*")])
    assert sorted(generate_site) == sorted(PLUGIN_USE_DIRECTORY_URLS)


def test_plugin_use_directory_urls_static_nav():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "search": {},
            "i18n": {
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
        nav=[
            {
                "Home": "index.md",
            }
        ],
    )
    build(mkdocs_config)
    site_dir = mkdocs_config["site_dir"]
    generate_site = [f.relative_to(site_dir) for f in Path(site_dir).glob("**/*.html")]
    generate_site.extend([f.relative_to(site_dir) for f in Path(site_dir).glob("**/image*.*")])
    assert sorted(generate_site) == sorted(PLUGIN_USE_DIRECTORY_URLS)


def test_plugin_no_use_directory_urls():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=False,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "search": {},
            "i18n": {
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
    )
    build(mkdocs_config)
    site_dir = mkdocs_config["site_dir"]
    generate_site = [f.relative_to(site_dir) for f in Path(site_dir).glob("**/*.html")]
    generate_site.extend([f.relative_to(site_dir) for f in Path(site_dir).glob("**/image*.*")])
    assert sorted(generate_site) == sorted(PLUGIN_NO_USE_DIRECTORY_URLS)


PLUGIN_USE_DIRECTORY_URLS_DEFAULT_ONLY = [
    Path("404.html"),
    Path("image.png"),
    Path("image.fake"),
    Path("index.html"),
    Path("assets/image_non_localized.png"),
    Path("english_default/index.html"),
    Path("topic1/named_file/index.html"),
    Path("topic2/index.html"),
    Path("topic2/1.1.filename.html"),
    Path("topic2/release_notes_17.1/index.html"),
    Path("topic2/release_notes_17.2/index.html"),
]
PLUGIN_NO_USE_DIRECTORY_URLS_DEFAULT_ONLY = [
    Path("404.html"),
    Path("image.png"),
    Path("image.fake"),
    Path("index.html"),
    Path("assets/image_non_localized.png"),
    Path("english_default/index.html"),
    Path("topic1/named_file.html"),
    Path("topic2/index.html"),
    Path("topic2/1.1.filename.html"),
    Path("topic2/release_notes_17.1.html"),
    Path("topic2/release_notes_17.2.html"),
]


def test_plugin_use_directory_urls_default_language_only():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "search": {},
            "i18n": {
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français", "build": False},
                ],
            },
        },
    )
    build(mkdocs_config)
    site_dir = mkdocs_config["site_dir"]
    generate_site = [f.relative_to(site_dir) for f in Path(site_dir).glob("**/*.html")]
    generate_site.extend([f.relative_to(site_dir) for f in Path(site_dir).glob("**/image*.*")])
    assert sorted(generate_site) == sorted(PLUGIN_USE_DIRECTORY_URLS_DEFAULT_ONLY)


def test_plugin_no_use_directory_urls_default_language_only():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=False,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "search": {},
            "i18n": {
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français", "build": False},
                ],
            },
        },
    )
    build(mkdocs_config)
    site_dir = mkdocs_config["site_dir"]
    generate_site = [f.relative_to(site_dir) for f in Path(site_dir).glob("**/*.html")]
    generate_site.extend([f.relative_to(site_dir) for f in Path(site_dir).glob("**/image*.*")])
    assert sorted(generate_site) == sorted(PLUGIN_NO_USE_DIRECTORY_URLS_DEFAULT_ONLY)



from collections import defaultdict
from copy import deepcopy
from pathlib import Path, PurePath
from typing import Union
from urllib.parse import urlsplit

from mkdocs import localization
from mkdocs.config.base import LegacyConfig
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin, get_plugin_logger
from mkdocs.structure.files import File, Files
from mkdocs.structure.nav import Navigation
from mkdocs.structure.pages import Page
from mkdocs.theme import Theme

from mkdocs_static_i18n import __file__ as installation_path
from mkdocs_static_i18n import folder, is_relative_to, suffix
from mkdocs_static_i18n.config import I18nPluginConfig

log = get_plugin_logger(__name__)

try:
    from importlib.metadata import version

    MATERIAL_VERSION = version("mkdocs-material")
except Exception:
    try:
        # python 3.7 compatibility, drop on 3.7 EOL
        import pkg_resources

        MATERIAL_VERSION = pkg_resources.get_distribution("mkdocs-material").version
    except Exception:
        MATERIAL_VERSION = None

try:
    from importlib.metadata import files as package_files

    LUNR_LANGUAGES = [
        PurePath(lang.stem).suffix.replace(".", "")
        for lang in package_files("mkdocs")
        if "mkdocs/contrib/search/lunr-language/lunr." in lang.as_posix() and len(lang.stem) == 7
    ]
    assert len(LUNR_LANGUAGES) > 1
    LUNR_LANGUAGES.append("en")
except Exception:
    LUNR_LANGUAGES = [
        "ar",
        "da",
        "de",
        "du",
        "en",
        "es",
        "fi",
        "fr",
        "hi",
        "hu",
        "hy",
        "it",
        "ja",
        "jp",
        "kn",
        "ko",
        "nl",
        "no",
        "pt",
        "ro",
        "ru",
        "sa",
        "sv",
        "ta",
        "te",
        "th",
        "tr",
        "vi",
        "zh",
    ]
    log.warning("Unable to detect lunr languages from mkdocs distribution")
MKDOCS_THEMES = ["mkdocs", "readthedocs"]


class ExtendedPlugin(BasePlugin[I18nPluginConfig]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.building = False
        self.current_language = None
        self.extra_alternate = {}
        self.i18n_files_per_language = {}
        self.original_configs = {}
        self.original_theme_configs = {}
        self.search_entries = []

    @property
    def all_languages(self):
        return [lang.locale for lang in self.config.languages]

    @property
    def default_language(self):
        for lang_config in self.config.languages:
            if lang_config.default is True:
                return lang_config.locale

    @property
    def current_language_config(self):
        return self.get_language_config(self.current_language)

    @property
    def is_default_language_build(self):
        return self.current_language == self.default_language

    @property
    def build_languages(self):
        return [
            lang.locale for lang in filter(lambda lang: lang.build is True, self.config.languages)
        ]

    def get_language_config(self, locale):
        for lang_config in filter(lambda lang: lang.locale == locale, self.config.languages):
            return lang_config
        raise Exception(f"Could not find language locale '{locale}'")

    def reconfigure_mkdocs_config(self, config: MkDocsConfig) -> MkDocsConfig:
        # MkDocs themes specific reconfiguration
        if config.theme.name in MKDOCS_THEMES:
            self.reconfigure_mkdocs_theme(config, self.current_language)

        # material theme specific reconfiguration (can be disabled)
        if config.theme.name == "material" and self.config.reconfigure_material is True:
            # check and warn about missing mkdocs-material version
            if MATERIAL_VERSION is None:
                log.warning("mkdocs-material version could not be detected")
            elif MATERIAL_VERSION <= "7.1":
                log.warning(f"mkdocs-material version {MATERIAL_VERSION} is not supported")
            else:
                config = self.reconfigure_material_theme(config, self.current_language)
                # warn about navigation.instant incompatibility
                if "navigation.instant" in config.theme._vars.get("features", []):
                    log.warning(
                        "mkdocs-material language switcher contextual link is not "
                        "compatible with theme.features = navigation.instant"
                    )

        # supported plugins reconfiguration
        for name, plugin in config.plugins.items():
            # search plugin (MkDocs & material > 9.0) reconfiguration
            if name in ["search", "material/search"]:
                # search plugin reconfiguration can be disabled
                if self.config.reconfigure_search:
                    config = self.reconfigure_search_plugin(config, name, plugin)
            if name == "with-pdf":
                config = self.reconfigure_with_pdf_plugin(config)

        # apply localized user config overrides
        config = self.apply_user_overrides(config)

        # Install a i18n aware version of sitemap.xml if not provided by the user
        if not Path(
            PurePath(config.theme._vars.get("custom_dir", ".")) / PurePath("sitemap.xml")
        ).exists():
            custom_i18n_sitemap_dir = Path(
                PurePath(installation_path).parent / PurePath("custom_i18n_sitemap")
            ).resolve()
            config.theme.dirs.insert(0, str(custom_i18n_sitemap_dir))

        return config

    def reset_to_original_config(self, config: MkDocsConfig):
        """
        Since applying user overrides on each language build is done
        on the same MkDocs config instance, we need to reset parts of
        it before overriding it again.

        Example:
            - the general config has a default site_name defined
            - we build for "en", "fr", "de" (in that order)
            - if site_name is overridden for "fr" but NOT for "de"
            - then "de" would get the "fr" site_name instead of the general one

        This method takes care of that problem.
        """
        for config_key, config_value in self.original_configs.items():
            config[config_key] = config_value
        for config_key, config_value in self.original_theme_configs.items():
            config.theme._vars[config_key] = config_value

        return config

    def save_original_config(self, store, key, value):
        if key not in store:
            store[key] = deepcopy(value)

    def apply_user_overrides(self, config: MkDocsConfig):
        """
        The i18n configuration structure allows users to set abitrary configuration
        that will be overriden if they match valid MkDocsConfig or Theme options.
        """
        # reset config to its orginal values since the config might have been
        # altered by a previous build
        config = self.reset_to_original_config(config)

        # some config overrides are forbidden as they make no sense
        forbidden_config_overrides = [
            "dev_addr",
            "docs_dir",
            "edit_uri_template",
            "edit_uri",
            "exclude_docs",
            "extra_css",
            "extra_javascript",
            "extra_templates",
            "hooks",
            "markdown_extensions",
            "mdx_configs",
            "not_in_nav",
            "plugins",
            "remote_branch",
            "remote_name",
            "repo_name",
            "repo_url",
            "site_dir",
            "strict",
            "use_directory_urls",
            "validation",
            "watch",
        ]
        for lang_key, lang_override in self.current_language_config.items():
            if lang_key in forbidden_config_overrides:
                log.warning(
                    f"Ignoring forbidden '{self.current_language}' config override '{lang_key}'"
                )
                continue
            if lang_key in config.data and lang_override is not None:
                mkdocs_config_option_type = type(config.data[lang_key])
                # support special Theme object overrides
                if mkdocs_config_option_type == Theme and isinstance(lang_override, dict):
                    config.theme = self.apply_user_theme_overrides(config.theme, lang_override)
                elif mkdocs_config_option_type in [str, bool, dict, list, type(None)]:
                    self.save_original_config(
                        self.original_configs, lang_key, config.data[lang_key]
                    )
                    config.load_dict({lang_key: lang_override})
                    log.info(
                        f"Overriding '{self.current_language}' config '{lang_key}' with '{lang_override}'"
                    )
                # extra is a legacy dict config
                elif mkdocs_config_option_type == LegacyConfig:
                    self.save_original_config(
                        self.original_configs, lang_key, config.data[lang_key]
                    )
                    config[lang_key].update(lang_override)
                    log.info(
                        f"Updating '{self.current_language}' config '{lang_key}' with '{lang_override}'"
                    )
                else:
                    log.warning(f"Unknown '{self.current_language}' config override '{lang_key}'")
        return config

    def apply_user_theme_overrides(self, theme: Theme, options: dict) -> Theme:
        """
        Support and apply special mkdocs.Theme object overrides.
        """

        def dict_recursive_update(source, overrides):
            for key, value in overrides.items():
                if isinstance(value, dict) and value:
                    updated = dict_recursive_update(source.get(key, {}), value)
                    source[key] = updated
                else:
                    source[key] = overrides[key]
            return source

        for key, value in options.items():
            if key in theme._vars and type(theme._vars[key]) == type(value):
                self.save_original_config(self.original_theme_configs, key, theme._vars[key])
                log.info(
                    f"Overriding '{self.current_language}' config 'theme.{key}' with '{value}'"
                )
                if isinstance(value, dict):
                    dict_recursive_update(theme._vars[key], value)
                elif isinstance(value, list):
                    for idx, item in enumerate(value):
                        if isinstance(item, dict):
                            dict_recursive_update(theme._vars[key][idx], item)
                        else:
                            theme._vars[key][idx] = item
                else:
                    theme._vars[key] = value
            elif key == "locale":
                self.save_original_config(self.original_theme_configs, key, theme._vars[key])
                theme._vars[key] = localization.parse_locale(value)
                log.info(
                    f"Overriding '{self.current_language}' config 'theme.{key}' with '{value}'"
                )
            else:
                log.warning(f"Unknown '{self.current_language}' config override 'theme.{key}'")
        return theme

    def reconfigure_mkdocs_theme(self, config: MkDocsConfig, locale: str) -> Theme:
        # set theme locale
        if "locale" in config.theme._vars:
            config.theme._vars["locale"] = localization.parse_locale(locale)

    def reconfigure_material_theme(self, config: MkDocsConfig, locale: str) -> MkDocsConfig:
        # set theme language
        if "language" in config.theme._vars:
            config.theme._vars["language"] = locale
        # configure extra.alternate language switcher
        if len(self.build_languages) > 1 or "null" in self.all_languages:
            # 'on_page_context' overrides the config.extra.alternate
            # so we need to reset it to its initial computed value if present
            if self.extra_alternate:
                config.extra["alternate"] = deepcopy(self.extra_alternate)
            # user has setup its own extra.alternate
            # warn if it's poorly configured
            if "alternate" in config.extra:
                for alternate in config.extra["alternate"]:
                    alternate_link = alternate.get("link", "")
                    if not alternate_link.startswith("http") and (
                        not alternate_link.startswith("/") or not alternate_link.endswith("/")
                    ):
                        log.info(
                            "The 'extra.alternate' configuration contains a "
                            "'link' option that should be an absolute path '/' and "
                            f"end with a trailing '/' in {alternate}"
                        )
                    for required_key in ["name", "link", "lang"]:
                        if required_key not in alternate:
                            log.info(
                                "The 'extra.alternate' configuration is missing a required "
                                f"'{required_key}' option in {alternate}"
                            )
            # configure the extra.alternate for the user
            # https://squidfunk.github.io/mkdocs-material/setup/changing-the-language/#site-language-selector
            else:
                if config.site_url:
                    base_url = urlsplit(config.site_url).path.rstrip("/")
                else:
                    base_url = ""
                config.extra["alternate"] = []
                # Add index.html file name when used with
                # use_directory_urls = True
                link_suffix = "" if config.get("use_directory_urls") else "index.html"
                # setup language switcher
                for language in self.all_languages:
                    lang_config = self.get_language_config(language)
                    # skip language if not built unless we are the special "null" locale
                    if lang_config.build is False and lang_config.locale != "null":
                        continue
                    config.extra["alternate"].append(
                        {
                            "name": lang_config.name,
                            "link": lang_config.fixed_link
                            or f"{base_url}{lang_config.link}{link_suffix}",
                            "lang": language,
                        }
                    )
            self.extra_alternate = deepcopy(config.extra["alternate"])
        return config

    def reconfigure_search_plugin(
        self, config: MkDocsConfig, search_plugin_name: str, search_plugin
    ):
        search_langs = search_plugin.config.lang or []
        for language in self.build_languages:
            if language in LUNR_LANGUAGES:
                if language not in search_langs:
                    search_langs.append(language)
                    log.info(
                        f"Adding '{language}' to the '{search_plugin_name}' plugin 'lang' option"
                    )
            elif language == self.current_language:
                log.info(
                    f"Language '{language}' is not supported by "
                    f"lunr.js, not setting it in the 'plugins.search.lang' option"
                )
        if search_langs:
            search_plugin.config.lang = search_langs
        return config

    def reconfigure_with_pdf_plugin(self, config: MkDocsConfig):
        """
        Support plugin mkdocs-with-pdf, see #110.
        """
        for events in config.plugins.events.values():
            for idx, event in enumerate(list(events)):
                try:
                    if str(event.__module__) == "mkdocs_with_pdf.plugin":
                        events.pop(idx)
                except AttributeError:
                    # partials don't have a module
                    pass
        return config

    def reconfigure_navigation(
        self,
        nav: Navigation,
        config: MkDocsConfig,
        files: Union[suffix.I18nFiles, folder.I18nFiles],
        nav_helper,
    ):
        """
        Apply static navigation items translation mapping for the current language.

        Localize the default homepage button.
        """
        # nav_translations
        nav_translations = self.current_language_config.nav_translations or {}

        for item in nav:
            if hasattr(item, "title") and item.title in nav_translations:
                item.title = nav_translations[item.title]
                nav_helper.translated_items += 1

            # is that the localized content homepage?
            if nav_helper.homepage is None and isinstance(item, Page):
                if item.url in nav_helper.expected_homepage_urls or item.url == "":
                    nav_helper.homepage = item

            # translation should be recursive to children
            if hasattr(item, "children") and item.children:
                self.reconfigure_navigation(item.children, config, files, nav_helper)

        return nav

    def reconfigure_page_context(self, context, page, config: MkDocsConfig, nav: Navigation):
        """
        Support dynamic reconfiguration of the material language selector so that
        users can switch between the different localized versions of their current page.
        """
        if self.extra_alternate:
            config.extra.alternate = deepcopy(self.extra_alternate)
            if PurePath(page.url) == PurePath("."):
                return context
            if PurePath(page.url) == PurePath(page.file.locale_alternate_of):
                return context
            #
            for extra_alternate in config.extra.alternate:
                alternate_lang = extra_alternate["lang"]
                # current page has an alternate for this language, use it
                if alternate_lang in page.file.alternates:
                    alternate_file = page.file.alternates[alternate_lang]
                    extra_alternate["link"] = alternate_file.url
        return context

    def extend_search_entries(self, config: MkDocsConfig):
        """
        Stack up search plugin entries as we build languages one after the other
        and deduplicate entries at the same time.
        """
        for name, plugin in config.plugins.items():
            if name in ["search", "material/search"]:
                if hasattr(plugin, "search_index"):
                    for entries_attr in ["entries", "_entries"]:
                        if hasattr(plugin.search_index, entries_attr):
                            entries = getattr(plugin.search_index, entries_attr)
                            self.search_entries.extend(entries)
                            break
                    else:
                        log.warning(f"Could not find '{name}' plugin search entries")

    def reconfigure_search_duplicates(self, search_index_entries):
        """
        We want to avoid indexing the same pages twice if the default language
        has its own version built as well as the /language version too as this
        would pollute the search results.
        When this happens, we favor the default language location if its
        content is the same as its /language counterpart.
        """
        default_lang_entries = filter(
            lambda x: not x["location"].startswith(
                tuple([f"{lang}/" for lang in self.build_languages])
            ),
            search_index_entries,
        )
        target_lang_entries = list(
            filter(
                lambda x: x["location"].startswith(
                    tuple([f"{lang}/" for lang in self.build_languages])
                ),
                search_index_entries,
            )
        )
        for default_lang_entry in default_lang_entries:
            duplicated_entries = filter(
                lambda x: x["title"] == default_lang_entry["title"]
                and x["location"].endswith(default_lang_entry["location"])
                and x["text"] == default_lang_entry["text"],
                target_lang_entries,
            )
            for duplicated_entry in duplicated_entries:
                log.debug(
                    f"removed duplicated search entry: {duplicated_entry['title']} "
                    f"{duplicated_entry['location']}"
                )
                search_index_entries.remove(duplicated_entry)
                target_lang_entries.remove(duplicated_entry)

    def reconfigure_search_index(self, config: MkDocsConfig):
        """
        Stack up search plugin entries as we build languages one after the other
        and deduplicate entries at the same time.
        """
        for name, plugin in config.plugins.items():
            if name in ["search", "material/search"]:
                attribute_name = (
                    "_entries" if hasattr(plugin.search_index, "_entries") else "entries"
                )
                try:
                    search_index_entries = getattr(plugin.search_index, attribute_name)
                except AttributeError:
                    log.warning(
                        f"Can't access the search index entries in {name} ({attribute_name})."
                    )
                    return
                # clear and repopulate the search index
                search_index_entries.clear()
                search_index_entries.extend(self.search_entries)
                # remove search index duplicates
                if self.config.reconfigure_search:
                    self.reconfigure_search_duplicates(search_index_entries)
                # run the post_build event to rebuild the search index
                plugin.on_post_build(config=config)

    def reconfigure_files(
        self,
        files: Files,
        mkdocs_config: MkDocsConfig,
    ) -> Union[suffix.I18nFiles, folder.I18nFiles]:
        """ """
        if self.config.docs_structure == "suffix":
            create_i18n_file = suffix.create_i18n_file
            I18nFiles = suffix.I18nFiles
        else:
            create_i18n_file = folder.create_i18n_file
            I18nFiles = folder.I18nFiles
        i18n_dest_uris = {}
        i18n_files = I18nFiles(self, [])
        i18n_alternate_dest_uris = defaultdict(list)
        for file in files:
            # user provided files in docs_dir
            if is_relative_to(file.abs_src_path, mkdocs_config.docs_dir):
                i18n_file = create_i18n_file(
                    file,
                    self.current_language,
                    self.default_language,
                    self.all_languages,
                    mkdocs_config,
                )

                # user provided documentation page
                if i18n_file.is_documentation_page():
                    # never seen that file?
                    if i18n_file.dest_uri not in i18n_dest_uris:
                        # best case scenario
                        # use the file since its locale is our current build language
                        if i18n_file.locale == self.current_language:
                            i18n_dest_uris[i18n_file.dest_uri] = i18n_file
                            log.debug(
                                f"Use {i18n_file.locale} {i18n_file.localization} {i18n_file}"
                            )
                        # if locale is the default language AND default language fallback is enabled
                        # we are using a file that is not really our locale
                        elif (
                            self.config.fallback_to_default is True
                            and i18n_file.locale == self.default_language
                        ):
                            i18n_dest_uris[i18n_file.dest_uri] = i18n_file
                            log.debug(
                                f"Use default {i18n_file.locale} {i18n_file.localization} {i18n_file}"
                            )
                            i18n_alternate_dest_uris[i18n_file.dest_uri].append(file)
                        else:
                            log.debug(
                                f"Ignore {i18n_file.locale} {i18n_file.localization} {i18n_file}"
                            )
                            i18n_alternate_dest_uris[i18n_file.dest_uri].append(file)

                    # we've seen that file already
                    else:
                        # override it only if this is our language
                        if i18n_file.locale == self.current_language:
                            # users should not add default non suffixed/folder files + suffixed/folder
                            # files when multiple languages are configured
                            if (
                                len(self.build_languages) > 1
                                and i18n_file.localization is not None
                                and i18n_dest_uris[i18n_file.dest_uri].locale == i18n_file.locale
                            ):
                                raise Exception(
                                    f"Conflicting files for the default language '{self.default_language}': "
                                    f"choose either '{i18n_file.src_uri}' or "
                                    f"'{i18n_dest_uris[i18n_file.dest_uri].src_uri}' but not both"
                                )
                            i18n_dest_uris[i18n_file.dest_uri] = i18n_file
                            log.debug(
                                f"Use localized {i18n_file.locale} {i18n_file.localization} {i18n_file}"
                            )
                        else:
                            log.debug(
                                f"Ignore {i18n_file.locale} {i18n_file.localization} {i18n_file}"
                            )
                            i18n_alternate_dest_uris[i18n_file.dest_uri].append(file)

                # user provided asset
                else:
                    # never seen that file?
                    if i18n_file.dest_uri not in i18n_dest_uris:
                        # best case scenario
                        # use the file since its locale is our current build language
                        if i18n_file.locale == self.current_language:
                            i18n_dest_uris[i18n_file.dest_uri] = i18n_file
                            log.debug(
                                f"Use asset {i18n_file.locale} {i18n_file.localization} {i18n_file}"
                            )
                        # if locale is the default language AND default language fallback is enabled
                        # we are using a file that is not really our locale
                        elif (
                            self.config.fallback_to_default is True
                            and i18n_file.locale == self.default_language
                        ):
                            i18n_asset = create_i18n_file(
                                file,
                                self.default_language,
                                self.default_language,
                                self.all_languages,
                                mkdocs_config,
                            )
                            i18n_dest_uris[i18n_file.dest_uri] = i18n_asset
                            log.debug(
                                f"Use asset default {i18n_asset.locale} {i18n_file.localization} {i18n_asset}"
                            )

                    # we've seen that file already
                    else:
                        # override it only if this is our language
                        if i18n_file.localization == self.current_language:
                            i18n_dest_uris[i18n_file.dest_uri] = i18n_file
                            log.debug(
                                f"Use asset localized {i18n_file.locale} {i18n_file.localization} {i18n_file}"
                            )

            # theme (and overrides) files
            elif not file.is_documentation_page():
                i18n_files.append(file)
            else:
                log.warning(f"Unhandled file case - {file.src_uri}")

        # populate the resulting Files and keep track of all the alternates
        # that will be used by the sitemap.xml template
        for file in i18n_dest_uris.values():
            log.debug(f"Selected {file.locale} {file.localization} {file}")
            i18n_files.append(file)

        # build the alternates for all the Files
        self.reconfigure_files_alternates(
            i18n_dest_uris, i18n_alternate_dest_uris, mkdocs_config, create_i18n_file
        )

        return i18n_files

    def reconfigure_files_alternates(
        self,
        i18n_dest_uris,
        i18n_alternate_dest_uris,
        mkdocs_config: MkDocsConfig,
        create_i18n_file: Union[suffix.create_i18n_file, folder.create_i18n_file],
    ):
        """
        Find and update the alternates of each file.
        """
        for build_lang in sorted(self.build_languages):
            for i18n_dest_uri, i18n_file in i18n_dest_uris.items():
                if build_lang not in i18n_file.alternates:
                    for alternate_file in i18n_alternate_dest_uris.get(i18n_dest_uri, []):
                        alternate_file = create_i18n_file(
                            alternate_file,
                            build_lang,
                            self.default_language,
                            self.all_languages,
                            mkdocs_config,
                        )
                        if alternate_file.locale == build_lang:
                            i18n_file.alternates[alternate_file.locale] = alternate_file
                            break
                    else:
                        # if fallbacking to default language is configured and we did not find
                        # an alternate for the build language, look for the default version of the file
                        if self.config.fallback_to_default is True:
                            for alternate_file in i18n_alternate_dest_uris.get(i18n_dest_uri, []):
                                alternate_file = create_i18n_file(
                                    alternate_file,
                                    build_lang,
                                    self.default_language,
                                    self.all_languages,
                                    mkdocs_config,
                                )
                                if alternate_file.locale == self.default_language:
                                    i18n_file.alternates[build_lang] = alternate_file
                                    break
                            else:
                                alternate_file = create_i18n_file(
                                    i18n_file,
                                    build_lang,
                                    self.default_language,
                                    self.all_languages,
                                    mkdocs_config,
                                )
                                if alternate_file.locale == self.default_language:
                                    i18n_file.alternates[build_lang] = alternate_file
        # uncomment to debug alternate selection
        # for i18n_dest_uri, i18n_file in i18n_dest_uris.items():
        #     print(" ")
        #     print(f"{i18n_dest_uri=}")
        #     for build_lang, alternate_file in i18n_file.alternates.items():
        #         print(
        #             f"    {build_lang=} {alternate_file.src_uri=} {alternate_file.locale_alternate_of=} {alternate_file.dest_uri=}"
        #         )

    def reconfigure_material_blog(self, nav: Navigation, mkdocs_config: MkDocsConfig, files: Files):
        """
        Since the material/blog plugin does modify the files structure and the navigation
        we need to override them and rebuild the blog/ part of it ourselves for now.

        See: https://github.com/squidfunk/mkdocs-material/issues/5909
        """
        if self.config.docs_structure == "suffix":
            create_i18n_file = suffix.create_i18n_file
        else:
            create_i18n_file = folder.create_i18n_file

        i18n_dest_uris = {}
        i18n_alternate_dest_uris = defaultdict(list)

        # at this point, the files have been cleaned up contrary to the
        # initial on_files run so it's harder to build the alternates
        for file in files:
            if is_relative_to(file.url, "blog/"):
                # get a localized version of the blog file
                i18n_file = create_i18n_file(
                    file,
                    self.current_language,
                    self.default_language,
                    self.all_languages,
                    mkdocs_config,
                )
                # used to rebuild blog alternates for the sitemap.xml and language switcher
                i18n_dest_uris[i18n_file.dest_uri] = file
                i18n_alternate_dest_uris[i18n_file.dest_uri].append(
                    File(
                        file.src_path,
                        mkdocs_config.docs_dir,
                        file.dest_path,
                        mkdocs_config.use_directory_urls,
                        dest_uri=file.dest_uri,
                        inclusion=file.inclusion,
                    )
                )
                # override blog related file specific properties in place
                file.abs_dest_path = i18n_file.abs_dest_path
                file.alternates = {self.current_language: file}
                file.dest_uri = i18n_file.dest_uri
                file.locale = i18n_file.locale
                file.locale_alternate_of = self.current_language
                file.url = i18n_file._get_url(mkdocs_config.use_directory_urls)
                #
                file.page._set_canonical_url(mkdocs_config.get('site_url', None))

        # reconfigure blog files alternates to the best we can
        # since we are past the on_files event and files useless to the current language
        # have been filtered out and new ones got generated by the blog plugin,
        # we can't get the complete view of the alternates
        self.reconfigure_files_alternates(
            i18n_dest_uris, i18n_alternate_dest_uris, mkdocs_config, create_i18n_file
        )

        # update the per-language files store for template alternates
        self.i18n_files_per_language[self.current_language] = files.documentation_pages()

        return nav



import pytest
from mkdocs.config.base import load_config
from mkdocs.structure.files import get_files
from mkdocs.structure.nav import get_navigation

STATIC_NAV_DIRECTORY_URLS = {
    "en": {"Home": "/", "Named File": "/topic1/named_file/", "Topic2": "/topic2/"},
    "fr": {"Home": "/fr/", "Named File": "/fr/topic1/named_file/", "Topic2": "/fr/topic2/"},
}
STATIC_NAV_NO_DIRECTORY_URLS = {
    "en": {
        "Home": "/index.html",
        "Named File": "/topic1/named_file.html",
        "Topic2": "/topic2/index.html",
    },
    "fr": {
        "Home": "/fr/index.html",
        "Named File": "/fr/topic1/named_file.html",
        "Topic2": "/fr/topic2/index.html",
    },
}
TRANSLATED_NAV_DIRECTORY_URLS = {
    "en": {"The Home": "/", "Renamed File": "/topic1/named_file/", "Topic2": "/topic2/"},
    "fr": {"Accueil": "/fr/", "Fichier Nommé": "/fr/topic1/named_file/", "Topic2": "/fr/topic2/"},
}
TRANSLATED_NAV_NO_DIRECTORY_URLS = {
    "en": {
        "The Home": "/index.html",
        "Renamed File": "/topic1/named_file.html",
        "Topic2": "/topic2/index.html",
    },
    "fr": {
        "Accueil": "/fr/index.html",
        "Fichier Nommé": "/fr/topic1/named_file.html",
        "Topic2": "/fr/topic2/index.html",
    },
}


@pytest.mark.parametrize(
    "plugin_config,use_directory_urls,control_data",
    [
        (
            {
                "docs_structure": "suffix",
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                ],
            },
            True,
            STATIC_NAV_DIRECTORY_URLS,
        ),
        (
            {
                "docs_structure": "suffix",
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                ],
            },
            False,
            STATIC_NAV_NO_DIRECTORY_URLS,
        ),
        (
            {
                "docs_structure": "suffix",
                "languages": [
                    {
                        "locale": "en",
                        "default": True,
                        "name": "english",
                        "nav_translations": {
                            "Home": "The Home",
                            "Named File": "Renamed File",
                        },
                    },
                    {
                        "locale": "fr",
                        "name": "français",
                        "nav_translations": {
                            "Home": "Accueil",
                            "Named File": "Fichier Nommé",
                        },
                    },
                ],
            },
            True,
            TRANSLATED_NAV_DIRECTORY_URLS,
        ),
        (
            {
                "docs_structure": "suffix",
                "languages": [
                    {
                        "locale": "en",
                        "default": True,
                        "name": "english",
                        "nav_translations": {
                            "Home": "The Home",
                            "Named File": "Renamed File",
                        },
                    },
                    {
                        "locale": "fr",
                        "name": "français",
                        "nav_translations": {
                            "Home": "Accueil",
                            "Named File": "Fichier Nommé",
                        },
                    },
                ],
            },
            False,
            TRANSLATED_NAV_NO_DIRECTORY_URLS,
        ),
        (
            {
                "docs_structure": "folder",
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                ],
            },
            True,
            STATIC_NAV_DIRECTORY_URLS,
        ),
        (
            {
                "docs_structure": "folder",
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                ],
            },
            False,
            STATIC_NAV_NO_DIRECTORY_URLS,
        ),
        (
            {
                "docs_structure": "folder",
                "languages": [
                    {
                        "locale": "en",
                        "default": True,
                        "name": "english",
                        "nav_translations": {
                            "Home": "The Home",
                            "Named File": "Renamed File",
                        },
                    },
                    {
                        "locale": "fr",
                        "name": "français",
                        "nav_translations": {
                            "Home": "Accueil",
                            "Named File": "Fichier Nommé",
                        },
                    },
                ],
            },
            True,
            TRANSLATED_NAV_DIRECTORY_URLS,
        ),
        (
            {
                "docs_structure": "folder",
                "languages": [
                    {
                        "locale": "en",
                        "default": True,
                        "name": "english",
                        "nav_translations": {
                            "Home": "The Home",
                            "Named File": "Renamed File",
                        },
                    },
                    {
                        "locale": "fr",
                        "name": "français",
                        "nav_translations": {
                            "Home": "Accueil",
                            "Named File": "Fichier Nommé",
                        },
                    },
                ],
            },
            False,
            TRANSLATED_NAV_NO_DIRECTORY_URLS,
        ),
    ],
)
def test_plugin_navigation(plugin_config, use_directory_urls, control_data):
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        docs_dir=f"docs_{plugin_config['docs_structure']}_structure_two_languages/",
        use_directory_urls=use_directory_urls,
        nav=[
            {"Home": "index.md"},
            {"Topic1": [{"Named File": "topic1/named_file.md"}]},
            {"Topic2": "topic2/README.md"},
            {"External": "https://ultrabug.fr"},
        ],
        plugins={
            "i18n": plugin_config,
        },
    )
    i18n_plugin = mkdocs_config["plugins"]["i18n"]
    for language in i18n_plugin.all_languages:
        config = i18n_plugin.on_config(mkdocs_config)
        #
        i18n_plugin.current_language = language
        #
        files = get_files(config)
        files = i18n_plugin.on_files(files, config)
        nav = get_navigation(files, config)
        nav = i18n_plugin.on_nav(nav, config, files)
        assert len(mkdocs_config.nav) == len(nav.pages) + 1  # +1 for external link
        for page in nav.pages:
            assert page.title in control_data[language]
            assert page.abs_url == control_data[language][page.title]
        assert nav.homepage is not None



import pytest
from mkdocs.config.base import load_config
from mkdocs.exceptions import Abort


def test_plugin_languages_no_default():
    with pytest.raises(Abort):
        load_config(
            "tests/mkdocs.yml",
            theme={"name": "mkdocs"},
            use_directory_urls=True,
            docs_dir="docs_suffix_structure_one_language/",
            plugins={
                "i18n": {
                    "languages": [
                        {
                            "locale": "en",
                            "name": "english",
                            "default": False,
                        },
                    ],
                },
            },
        )


def test_plugin_languages_no_build():
    with pytest.raises(Abort):
        load_config(
            "tests/mkdocs.yml",
            theme={"name": "mkdocs"},
            use_directory_urls=True,
            docs_dir="docs_suffix_structure_two_languages/",
            plugins={
                "i18n": {
                    "languages": [
                        {
                            "locale": "en",
                            "name": "english",
                            "default": True,
                            "build": False,
                        },
                        {"locale": "fr", "name": "français", "build": False},
                    ],
                },
            },
        )


def test_plugin_languages_dual_lang():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "i18n": {
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
    )
    i18n_plugin = mkdocs_config["plugins"]["i18n"]
    i18n_plugin.on_config(mkdocs_config)
    assert i18n_plugin.config["languages"] == [
        {
            "locale": "en",
            "name": "english",
            "link": "/",
            "fixed_link": None,
            "build": True,
            "default": True,
            "nav": None,
            "nav_translations": None,
            "theme": None,
            "site_name": None,
            "copyright": None,
            "extra": None,
            "site_author": None,
            "site_description": None,
            "site_url": None,
        },
        {
            "locale": "fr",
            "name": "français",
            "link": "/fr/",
            "fixed_link": None,
            "build": True,
            "default": False,
            "nav": None,
            "nav_translations": None,
            "theme": None,
            "site_name": None,
            "copyright": None,
            "extra": None,
            "site_author": None,
            "site_description": None,
            "site_url": None,
        },
    ]


def test_plugin_languages_one_lang():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_one_language/",
        plugins={
            "i18n": {
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                ],
            },
        },
    )
    i18n_plugin = mkdocs_config["plugins"]["i18n"]
    i18n_plugin.on_config(mkdocs_config)
    assert i18n_plugin.config["languages"] == [
        {
            "locale": "en",
            "name": "english",
            "link": "/",
            "fixed_link": None,
            "build": True,
            "default": True,
            "nav": None,
            "nav_translations": None,
            "theme": None,
            "site_name": None,
            "copyright": None,
            "extra": None,
            "site_author": None,
            "site_description": None,
            "site_url": None,
        },
    ]


def test_plugin_languages_null_no_fixed_link():
    with pytest.raises(Abort):
        load_config(
            "tests/mkdocs.yml",
            theme={"name": "mkdocs"},
            use_directory_urls=True,
            docs_dir="docs_suffix_structure_two_languages/",
            plugins={
                "i18n": {
                    "languages": [
                        {
                            "locale": "en",
                            "name": "english",
                            "default": True,
                        },
                        {"locale": "fr", "name": "français"},
                        {"locale": "null", "name": "help"},
                    ],
                },
            },
        )


def test_plugin_languages_dual_lang_with_null():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "i18n": {
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                    {"locale": "null", "name": "help", "fixed_link": "https://ultrabug.fr"},
                ],
            },
        },
    )
    i18n_plugin = mkdocs_config["plugins"]["i18n"]
    i18n_plugin.on_config(mkdocs_config)
    assert i18n_plugin.config["languages"] == [
        {
            "locale": "en",
            "name": "english",
            "link": "/",
            "fixed_link": None,
            "build": True,
            "default": True,
            "nav": None,
            "nav_translations": None,
            "theme": None,
            "site_name": None,
            "copyright": None,
            "extra": None,
            "site_author": None,
            "site_description": None,
            "site_url": None,
        },
        {
            "locale": "fr",
            "name": "français",
            "link": "/fr/",
            "fixed_link": None,
            "build": True,
            "default": False,
            "nav": None,
            "nav_translations": None,
            "theme": None,
            "site_name": None,
            "copyright": None,
            "extra": None,
            "site_author": None,
            "site_description": None,
            "site_url": None,
        },
        {
            "locale": "null",
            "name": "help",
            "link": "/null/",
            "fixed_link": "https://ultrabug.fr",
            "build": False,
            "default": False,
            "nav": None,
            "nav_translations": None,
            "theme": None,
            "site_name": None,
            "copyright": None,
            "extra": None,
            "site_author": None,
            "site_description": None,
            "site_url": None,
        },
    ]


def value_in_frontmatter(key, metadata):
    """Check if a key exists in the frontmatter.

    Args:
        key (any): the key to check
        metadata (any): the frontmatter

    Returns:
        bool: true if exists
    """
    if key in metadata:
        return metadata[key]
    else:
        return None


def on_env(env, config, files, **kwargs):
    env.filters["value_in_frontmatter"] = value_in_frontmatter
    return env



from pathlib import Path

from mkdocs.commands.build import build
from mkdocs.config.base import load_config


def test_rtd_fontfiles():
    config_base_rtd = load_config(
        "tests/mkdocs.yml",
        theme={"name": "readthedocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
    )

    config_plugin_rtd = load_config(
        "tests/mkdocs.yml",
        theme={"name": "readthedocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "i18n": {
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
    )

    base_site_dir = config_base_rtd["site_dir"]
    i18n_site_dir = config_plugin_rtd["site_dir"]

    build(config_base_rtd)
    build(config_plugin_rtd)

    base_font_files = [
        str(p).replace(base_site_dir, "") for p in Path(base_site_dir, "fonts").glob("**/*")
    ]
    i18n_font_files = [
        str(p).replace(i18n_site_dir, "") for p in Path(i18n_site_dir, "fonts").glob("**/*")
    ]
    assert sorted(base_font_files) == sorted(i18n_font_files)



"""Utility functions and classes that aren't limited to any scenario."""
import logging
from typing import Dict, Optional, TypeVar

from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin

Plugin = TypeVar("Plugin", bound=BasePlugin)
"""Plugin Instance Type"""


def get_plugin(name: str, config: MkDocsConfig) -> Optional[Plugin]:
    """Returns a plugin instance"""

    plugins: Dict[str, BasePlugin] = config["plugins"]
    instance: Plugin = plugins.get(name)

    if instance:
        return instance

    scoped_name: str = f"/{name}"

    # If the plugin was not found using the name
    # then look for a theme-namespaced plugin.
    for n, p in plugins.items():
        if n.endswith(scoped_name):
            return p

    return None


class I18nLoggingFilter:
    """Avoid logging duplicate build time messages."""

    def __init__(self, *_, **__):
        self.filtered_prefixes = set()

    def __call__(self, record: logging.LogRecord) -> bool:
        for prefix in self.filtered_prefixes:
            if record.msg.startswith(prefix):
                return False

        return True



import logging
import sys
from pathlib import PurePath
from typing import Optional

from jinja2.ext import loopcontrols
from mkdocs import plugins
from mkdocs.commands.build import build
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import get_plugin_logger
from mkdocs.structure.files import Files
from mkdocs.structure.pages import Page

from mkdocs_static_i18n import folder
from mkdocs_static_i18n.reconfigure import ExtendedPlugin
from mkdocs_static_i18n.utils import I18nLoggingFilter

log = get_plugin_logger(__name__)


class I18n(ExtendedPlugin):
    """
    We use 'event_priority' to make sure that we are last plugin to be executed
    because we need to make sure that we react to other plugins' behavior
    properly.

    Current plugins we heard of and require that we control their order:
        - awesome-pages: this plugin should run before us
        - with-pdf: this plugin is triggerd by us on the appropriate on_* events
    """

    @plugins.event_priority(-100)
    def on_config(self, config: MkDocsConfig):
        """
        Enrich configuration with language specific knowledge.
        """
        # first execution, setup defaults
        if self.current_language is None:
            self.current_language = self.default_language

        path_suffix = self.current_language if not self.is_default_language_build else ""

        log.info(
            f"Building '{self.current_language}' documentation to directory: "
            f"{PurePath(config.site_dir) / path_suffix}"
        )

        # reconfigure the mkdocs config
        return self.reconfigure_mkdocs_config(config)

    @plugins.event_priority(-100)
    def on_files(self, files: Files, config: MkDocsConfig):
        """
        Construct the lang specific file tree which will be used to
        generate the navigation for the default site and per language.

        Note that each file's alternates are also built during this step.
        """
        i18n_files = self.reconfigure_files(files, config)
        # update the (cumulative) global alternates map which is
        # used by the sitemap.xml template
        self.i18n_files_per_language[self.current_language] = i18n_files.documentation_pages()
        return i18n_files

    @plugins.event_priority(-100)
    def on_nav(self, nav, config, files):
        """
        Translate i18n aware navigation to honor the 'nav_translations' option.
        """
        # folder structure, reconfigure navigation to remove language sections
        if self.config.docs_structure == "folder":
            nav = folder.reconfigure_navigation(self, nav)

        # reconfigure mkdocs-material blog navigation
        if self.config.reconfigure_material and "material/blog" in config.plugins:
            nav = self.reconfigure_material_blog(nav, config, files)

        homepage_suffix: str = "" if config.use_directory_urls else "index.html"

        class NavHelper:
            translated_items: int = 0
            homepage: Optional[Page] = nav.homepage
            expected_homepage_urls = [
                f"{self.current_language}/{homepage_suffix}",
                f"/{self.current_language}/{homepage_suffix}",
            ]

        i18n_nav = self.reconfigure_navigation(nav, config, files, NavHelper)
        i18n_nav.homepage = NavHelper.homepage

        # report translated entries
        if NavHelper.translated_items:
            log.info(
                f"Translated {NavHelper.translated_items} navigation element"
                f"{'s' if NavHelper.translated_items > 1 else ''} to '{self.current_language}'"
            )

        # report missing homepage
        if i18n_nav.homepage is None:
            log.warning(f"Could not find a homepage for locale '{self.current_language}'")

        # manually trigger with-pdf, see #110
        with_pdf_plugin = config.plugins.get("with-pdf")
        if with_pdf_plugin:
            with_pdf_plugin.on_nav(i18n_nav, config, files)

        return i18n_nav

    def on_env(self, env, config, files):
        # Add extension to allow the "continue" clause in the sitemap template loops.
        env.add_extension(loopcontrols)

    @plugins.event_priority(50)
    def on_template_context(self, context, template_name, config):
        """
        Template context only applies to Template() objects.
        We add some metadata for users and our sitemap.xml generation.
        """
        # convenience for users in case they need it (we don't)
        context["i18n_build_languages"] = self.build_languages
        context["i18n_current_language_config"] = self.current_language_config
        context["i18n_current_language"] = self.current_language
        # used by sitemap.xml template
        context["i18n_alternates"] = self.i18n_files_per_language
        return context

    @plugins.event_priority(50)
    def on_page_context(self, context, page, config, nav):
        """
        Page context only applies to Page() objects.
        We add some metadata for users as well as some neat reconfiguration features.

        Overriden templates such as the sitemap.xml are not impacted by this method!
        """
        if isinstance(page, Page) and hasattr(page.file, "locale"):
            # export some useful i18n related variables on page context, see #75
            context["i18n_config"] = self.config
            context["i18n_file_locale"] = page.file.locale
            context["i18n_page_locale"] = self.current_language
            if self.config.reconfigure_material is True:
                context = self.reconfigure_page_context(context, page, config, nav)
        return context

    @plugins.event_priority(-100)
    def on_post_page(self, output, page, config):
        """
        Some plugins we control ourselves need this event.
        """
        # manually trigger with-pdf, see #110
        with_pdf_plugin = config.plugins.get("with-pdf")
        if with_pdf_plugin:
            with_pdf_plugin.on_post_page(output, page, config)
        return output

    @plugins.event_priority(-100)
    def on_post_build(self, config):
        """
        We build every language on its own directory.
        """

        # memorize locale search entries
        self.extend_search_entries(config)

        if self.building:
            return

        self.building = True

        # Block time logging for internal builds and filter reduntant MkDocs log
        build_logger = logging.getLogger("mkdocs.commands.build")
        i18n_filter = I18nLoggingFilter()
        i18n_filter.filtered_prefixes.add("Documentation built in")
        i18n_filter.filtered_prefixes.add("Building documentation to directory")
        build_logger.addFilter(i18n_filter)

        # manually trigger with-pdf, see #110
        with_pdf_plugin = config.plugins.get("with-pdf")
        if with_pdf_plugin:
            with_pdf_output_path = with_pdf_plugin.config["output_path"]
            with_pdf_plugin.on_post_build(config)

        # monkey patching mkdocs.utils.clean_directory to avoid
        # the site_dir to be cleaned up on each build() call
        from mkdocs import utils

        mkdocs_utils_clean_directory = utils.clean_directory
        utils.clean_directory = lambda x: x

        for locale in self.build_languages:
            if locale == self.current_language:
                continue
            self.current_language = locale
            # TODO: reconfigure config here? skip on_config?
            dirty = True if "--dirty" in sys.argv or "--dirtyreload" in sys.argv else False
            build(config, dirty=dirty)

            # manually trigger with-pdf for this locale, see #110
            if with_pdf_plugin:
                with_pdf_plugin.config["output_path"] = PurePath(
                    f"{locale}/{with_pdf_output_path}"
                ).as_posix()
                with_pdf_plugin.on_post_build(config)

        # rebuild and deduplicate the search index
        self.reconfigure_search_index(config)

        # remove monkey patching in case some other builds are triggered
        # on the same site (tests, ci...)
        utils.clean_directory = mkdocs_utils_clean_directory

        # Unblock time logging after internal builds
        build_logger.removeFilter(i18n_filter)

        self.building = False



from mkdocs.commands.build import build
from mkdocs.config.base import load_config


def test_search_entries():
    for theme in ["mkdocs", "material"]:
        mkdocs_config = load_config(
            "tests/mkdocs.yml",
            theme={"name": theme},
            use_directory_urls=True,
            docs_dir="docs_suffix_structure_two_languages/",
            plugins={
                "search": {},
                "i18n": {
                    "languages": [
                        {
                            "locale": "en",
                            "name": "english",
                            "default": True,
                        },
                        {"locale": "fr", "name": "français"},
                    ],
                },
            },
        )
        if theme == "material":
            mkdocs_config["plugins"]["material/search"].on_startup(command=None, dirty=False)
        build(mkdocs_config)
        if theme == "mkdocs":
            search_plugin = mkdocs_config["plugins"]["search"]
            assert len(search_plugin.search_index._entries) == 36
        else:
            search_plugin = mkdocs_config["plugins"]["material/search"]
            assert len(search_plugin.search_index.entries) == 26


def test_search_entries_no_directory_urls():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=False,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "search": {},
            "i18n": {
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
    )
    build(mkdocs_config)
    search_plugin = mkdocs_config["plugins"]["search"]
    assert len(search_plugin.search_index._entries) == 36


def test_search_entries_no_reconfigure():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "search": {},
            "i18n": {
                "reconfigure_search": False,
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
    )
    build(mkdocs_config)
    search_plugin = mkdocs_config["plugins"]["search"]
    assert len(search_plugin.search_index._entries) == 40


def test_search_add_lang():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "search": {},
            "i18n": {
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
    )
    build(mkdocs_config)
    search_plugin = mkdocs_config["plugins"]["search"]
    assert search_plugin.config["lang"] == ["en", "fr"]


def test_search_add_missing_lang():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "search": {
                "lang": ["en"],
            },
            "i18n": {
                "reconfigure_search": True,
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
    )
    build(mkdocs_config)
    search_plugin = mkdocs_config["plugins"]["search"]
    assert search_plugin.config["lang"] == ["en", "fr"]


def test_search_no_add_lang():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "search": {
                "lang": ["en"],
            },
            "i18n": {
                "reconfigure_search": False,
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                    },
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
    )
    build(mkdocs_config)
    search_plugin = mkdocs_config["plugins"]["search"]
    assert search_plugin.config["lang"] == ["en"]



from mkdocs.commands.build import build
from mkdocs.config.base import load_config


def test_plugin_language_selector_use_directory_urls():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "material"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "i18n": {
                "languages": [
                    {"locale": "en", "name": "english", "default": True},
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
    )
    i18n_plugin = mkdocs_config["plugins"]["i18n"]
    result = i18n_plugin.on_config(mkdocs_config)
    assert result["extra"]["alternate"] == [
        {"name": "english", "link": "/", "lang": "en"},
        {"name": "français", "link": "/fr/", "lang": "fr"},
    ]


def test_plugin_language_selector_no_use_directory_urls():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "material"},
        use_directory_urls=False,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "i18n": {
                "languages": [
                    {"locale": "en", "name": "english", "default": True},
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
    )
    i18n_plugin = mkdocs_config["plugins"]["i18n"]
    result = i18n_plugin.on_config(mkdocs_config)
    assert result["extra"]["alternate"] == [
        {"name": "english", "link": "/index.html", "lang": "en"},
        {
            "name": "français",
            "link": "/fr/index.html",
            "lang": "fr",
        },
    ]


def test_plugin_language_selector_fixed_alternate():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "mkdocs"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        extra={
            "alternate": [
                {"name": "english", "link": "/default", "lang": "en"},
                {"name": "français", "link": "/french", "lang": "fr"},
            ]
        },
        plugins={
            "i18n": {
                "languages": [
                    {"locale": "en", "name": "english", "default": True},
                    {"locale": "fr", "name": "français"},
                ],
            },
        },
    )
    i18n_plugin = mkdocs_config["plugins"]["i18n"]
    result = i18n_plugin.on_config(mkdocs_config)
    assert result["extra"]["alternate"] == [
        {"name": "english", "link": "/default", "lang": "en"},
        {"name": "français", "link": "/french", "lang": "fr"},
    ]


def test_plugin_language_selector_single_default_language():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "material"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_one_language/",
        plugins={
            "i18n": {
                "languages": [{"locale": "fr", "name": "français", "default": True}],
            },
        },
    )
    i18n_plugin = mkdocs_config["plugins"]["i18n"]
    result = i18n_plugin.on_config(mkdocs_config)
    assert result["extra"] == {}


def test_plugin_language_selector_fixed_link():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "material"},
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "i18n": {
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                        "fixed_link": "/en",
                    },
                    {"locale": "fr", "name": "français", "fixed_link": "/fr"},
                ],
            }
        },
    )
    i18n_plugin = mkdocs_config["plugins"]["i18n"]
    result = i18n_plugin.on_config(mkdocs_config)
    assert result["extra"]["alternate"] == [
        {"name": "english", "link": "/en", "lang": "en"},
        {"name": "français", "link": "/fr", "lang": "fr"},
    ]
    build(mkdocs_config)


def test_plugin_language_selector_fixed_link_with_static_alternate():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "material"},
        docs_dir="docs_suffix_structure_two_languages/",
        extra={
            "alternate": [
                {"name": "english", "link": "/", "lang": "en"},
                {
                    "name": "français",
                    "link": "/fr/",
                    "lang": "fr",
                },
            ],
        },
        plugins={
            "i18n": {
                "languages": [
                    {
                        "locale": "en",
                        "name": "english",
                        "default": True,
                        "fixed_link": "/en",
                    },
                    {"locale": "fr", "name": "français", "fixed_link": "/fr"},
                ],
            }
        },
    )
    i18n_plugin = mkdocs_config["plugins"]["i18n"]
    result = i18n_plugin.on_config(mkdocs_config)
    assert result["extra"]["alternate"] == [
        {"name": "english", "link": "/", "lang": "en"},
        {"name": "français", "link": "/fr/", "lang": "fr"},
    ]
    build(mkdocs_config)


def test_plugin_language_selector_with_null():
    mkdocs_config = load_config(
        "tests/mkdocs.yml",
        theme={"name": "material"},
        use_directory_urls=True,
        docs_dir="docs_suffix_structure_two_languages/",
        plugins={
            "i18n": {
                "languages": [
                    {"locale": "en", "name": "english", "default": True},
                    {"locale": "fr", "name": "français"},
                    {"locale": "null", "name": "help", "fixed_link": "https://ultrabug.fr"},
                ],
            },
        },
    )
    i18n_plugin = mkdocs_config["plugins"]["i18n"]
    result = i18n_plugin.on_config(mkdocs_config)
    assert result["extra"]["alternate"] == [
        {"name": "english", "link": "/", "lang": "en"},
        {"name": "français", "link": "/fr/", "lang": "fr"},
        {"name": "help", "link": "https://ultrabug.fr", "lang": "null"}
    ]


from pathlib import PurePath

version = "1.1.1"


def is_relative_to(src_path, dest_path):
    try:
        return PurePath(src_path).is_relative_to(dest_path)
    except AttributeError:
        # python < 3.9 compat
        try:
            PurePath(src_path).relative_to(dest_path)
            return True
        except ValueError:
            return False
